
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <memory.h>
#include <assert.h>
#include <list>
#include <vector>
#include <bitset>
#include <math.h>
#include <time.h>
#include <signal.h>
#include <string>

using namespace std;

#include <Types.h>
#include <Utility.h>
#include <Random.h>

FILE *myfopen(cc8 * file, cc8 * mode) {
    FILE * ret = fopen(file, mode);
    if(!ret) {
        logPrintf("[ ERROR ]  Unable to open file: %s\n", file);
    }

    return ret;
}

FILE *mytmpfile() {
    FILE * ret = tmpfile();
    if(!ret) {
        logPrintf("[ ERROR ]  Unable to open temporary file!\n");
    }

    return ret;
}

//#define fopen(file, mode) myfopen(file, mode)
#define tmpfile() mytmpfile()


typedef struct vec2f {
    f32 x, y;
} PACKED vec2f;

typedef struct vec3f {
    vec3f() { }
    vec3f(f32 a, f32 b, f32 c) : x(a), y(b), z(c) { }

    f32 x, y, z;
} PACKED vec3f;

inline vec3f cross(const vec3f &a, const vec3f &b) {
    return vec3f((a.y * b.z) - (a.z * b.y), (a.z * b.x) - (a.x * b.z), (a.x * b.y) - (a.y * b.x));
}

inline void normalize(vec3f &v) {
    f32 oneOverLength = 1.0f / sqrtf(v.x * v.x + v.y * v.y + v.z * v.z);

    v.x *= oneOverLength;
    v.y *= oneOverLength;
    v.z *= oneOverLength;
}

inline vec3f subtract(const vec3f &a, const vec3f &b) {
    return vec3f(a.x - b.x, a.y - b.y, a.z - b.z);
}

inline vec3f add(const vec3f &a, const vec3f &b) {
    return vec3f(a.x + b.x, a.y + b.y, a.z + b.z);
}

/*
   vertex 3 (0, 1) |---------| vertex 2 (1, 1)
                   |         |
                   |         |
                   |         |
                   |         |
   vertex 0 (0, 0) |---------| vertex 1 (1, 0)
*/

// standardizes a the terrain blocks
typedef struct TerrainChunkStandard {
    vec2f min, max; // bounding area of the terrain chunk
    f32 edgeHeight; // height of the edges
    s32 maxResolution; // how many times can we subdevide a single chunk?
    f32 minHeight, maxHeight; // the min and max height of a vertex
} TerrainChunkStandard;

// a arbitrary-sized terrain chunk
typedef struct TerrainChunk {
    vec3f * vertices;
    vec2f * texcoords;
    u32 * indices;

    u8 * normalHeightMapPixels;

    s32 resolution;
} TerrainChunk;


static volatile bool done = false;


void sighandler(s32 sig) {
    done = true;
}

void saveTGA(char *file, s16 width, s16 height, u8 *outImage) {
    FILE *pFile = fopen(file, "wb");

    u8 uselessChar;
    s16 uselessInt;
    u8 imageType;
    s32 index;
    u8 bits;
    s32 Size;
    s32 colorMode;
    u8 tempColors;

    imageType = 2;
    colorMode = 4;
    bits = 32;
    uselessChar = 0;
    uselessInt = 0;

    fwrite(&uselessChar, sizeof(u8), 1, pFile);
    fwrite(&uselessChar, sizeof(u8), 1, pFile);
    fwrite(&imageType, sizeof(u8), 1, pFile);
    fwrite(&uselessInt, sizeof(s16), 1, pFile);
    fwrite(&uselessInt, sizeof(s16), 1, pFile);
    fwrite(&uselessChar, sizeof(u8), 1, pFile);
    fwrite(&uselessInt, sizeof(s16), 1, pFile);
    fwrite(&uselessInt, sizeof(s16), 1, pFile);
    fwrite(&width, sizeof(s16), 1, pFile);
    fwrite(&height, sizeof(s16), 1, pFile);
    fwrite(&bits, sizeof(u8), 1, pFile);
    fwrite(&uselessChar, sizeof(u8), 1, pFile);

    Size = width * height * colorMode;
    for(index = 0; index < Size; index += colorMode) {
        tempColors = outImage[index];
        outImage[index] = outImage[index + 2];
        outImage[index + 2] = tempColors;
    }

    fwrite(outImage, sizeof(u8), Size, pFile);
    fclose(pFile);
}

s32 main(s32 argc, c8 ** argv) {
    // seed the cryptographically secure random RNG
    srand(time(NULL));

    // setup signal handlers
/*
    signal(SIGABRT, sighandler);
    signal(SIGFPE, sighandler);
    signal(SIGILL, sighandler);
    signal(SIGINT, sighandler);
    signal(SIGSEGV, sighandler);
    signal(SIGTERM, sighandler);
*/

    // init the chunk standard
    TerrainChunkStandard * tStandard = new TerrainChunkStandard();

    tStandard->min.x = 0.0f;
    tStandard->min.y = 0.0f;

    f32 size = 1.0f + floorf((frand() * 10.0f) + 0.5f);
    tStandard->max.x = size;
    tStandard->max.y = size;

    tStandard->maxResolution = 10 + (rand() % 10);

    tStandard->edgeHeight = frandc(2);

    tStandard->minHeight = frand2();
    tStandard->maxHeight = tStandard->minHeight + frand();

    s32 chunkCount = 0;

    logPrintf("ODGen.\n");

    // generate a bunch of chunks (ctrl + c) to stop
    while(!__sync_lock_test_and_set(&done, false) && chunkCount < 512) {
        TerrainChunk * chunk = new TerrainChunk();

        chunk->resolution            = 2 + (rand() % tStandard->maxResolution);
        chunk->vertices              = new vec3f[chunk->resolution * chunk->resolution];
        chunk->texcoords             = new vec2f[chunk->resolution * chunk->resolution];
        chunk->indices               = new u32[(chunk->resolution - 1) * (chunk->resolution - 1) * 4];
        chunk->normalHeightMapPixels = new u8[chunk->resolution * chunk->resolution * 4];

        logPrintf("Chunk: %dx%d.\n", chunk->resolution, chunk->resolution);

        // loop through each chunk's tiles
        s32 x, y, row, index, qindex, rgbaIndex;
        for(y = 0; y < chunk->resolution; ++y) {
            row = y * chunk->resolution;
            for(x = 0; x < chunk->resolution; ++x) {
                index = x + row;

                vec3f &v  = chunk->vertices[index];
                vec2f &tc = chunk->texcoords[index];

                // calculate the texture coordinate
                tc.x = ((f32)x / (f32)chunk->resolution);
                tc.y = ((f32)y / (f32)chunk->resolution);

                // calculate the [x, z] coords (assuming +y is up)
                v.x = (tc.x * (tStandard->max.x - tStandard->min.x)) + tStandard->min.x;
                v.z = (tc.y * (tStandard->max.y - tStandard->min.y)) + tStandard->min.y;

                // if its a edge-vertex its height must match the standard's height (to ensure tiles sync up without seams)
                if(y == 0 || y == (chunk->resolution - 1) || x == 0 || x == (chunk->resolution - 1)) {
                    v.y = tStandard->edgeHeight;
                } else {
                    // otherwise its part of the inside-mesh, calculate the y coord randomly using the standard
                    v.y = (frand() * (tStandard->maxHeight - tStandard->minHeight)) + tStandard->minHeight;
                }

                // calculate quad indices
                if(x < (chunk->resolution - 1) && y < (chunk->resolution - 1)) {
                    qindex = ((y * (chunk->resolution - 1)) << 2) + (x << 2);

                    chunk->indices[qindex + 0] = ((y + 0) * chunk->resolution) + (x + 0) + 1; // [0, 0]
                    chunk->indices[qindex + 1] = ((y + 0) * chunk->resolution) + (x + 1) + 1; // [1, 0]
                    chunk->indices[qindex + 2] = ((y + 1) * chunk->resolution) + (x + 1) + 1; // [1, 1]
                    chunk->indices[qindex + 3] = ((y + 1) * chunk->resolution) + (x + 0) + 1; // [0, 1]
                }
            }
        }

        // calculate the normal-heightmap
/*
        for(y = 0; y < (chunk->resolution - 1); ++y) {
            for(x = 0; x < (chunk->resolution - 1); ++x) {
                // get the quad's vertices
                vec3f &v0 = chunk->vertices[(y + 0) * chunk->resolution + (x + 0)];
                vec3f &v1 = chunk->vertices[(y + 0) * chunk->resolution + (x + 1)];
                vec3f &v2 = chunk->vertices[(y + 1) * chunk->resolution + (x + 1)];
                vec3f &v3 = chunk->vertices[(y + 1) * chunk->resolution + (x + 0)];

                // calculate the two triangle normals
                vec3f normal0 = cross(subtract(v2, v0), subtract(v1, v0));
                normalize(normal0);

                vec3f normal1 = cross(subtract(v2, v0), subtract(v3, v0));
                normalize(normal1);

                // calculate the actual normal
                vec3f normal = add(normal0, normal1);
                normalize(normal);

                // set the pixels
                rgbaIndex = ((y * chunk->resolution) << 2) + (x << 2);

                chunk->normalHeightMapPixels[rgbaIndex + 0] = (u8)(normal.x * 255.0f);
                chunk->normalHeightMapPixels[rgbaIndex + 1] = (u8)(normal.y * 255.0f);
                chunk->normalHeightMapPixels[rgbaIndex + 2] = (u8)(normal.z * 255.0f);
                //chunk->normalHeightMapPixels[rgbaIndex + 3] = 255;
                chunk->normalHeightMapPixels[rgbaIndex + 3] = (u8)(((v0.y + v1.y + v2.y + v3.y) * 0.25f) * 255.0f);
            }
        }
*/

        // save the chunk as a wavefront *.obj model and normalHeightmap as a tga image
        c8 folderpath[256];
        sprintf(folderpath, "chunk_%d", chunkCount);
        mymkdir(folderpath);

        c8 filepath[256];
        strcpy(filepath, folderpath);
        strcat(filepath, "/model.obj");
        FILE * f = fopen(filepath, "wb");

/*
        c8 nhm_filepath[256];
        strcpy(nhm_filepath, folderpath);
        strcat(nhm_filepath, "/normalHeightmap.tga");
        saveTGA(nhm_filepath, chunk->resolution, chunk->resolution, chunk->normalHeightMapPixels);
*/

        // write the verts & texcoords
        u32 numElements = chunk->resolution * chunk->resolution;
        u32 i = 0;
        for(; i < numElements; ++i) {
            fprintf(f, "v %f %f %f\n", chunk->vertices[i].x, chunk->vertices[i].y, chunk->vertices[i].z);
            fprintf(f, "tc %f %f\n", chunk->texcoords[i].x, chunk->texcoords[i].y);
        }

        // write the indices
        numElements = (chunk->resolution - 1) * (chunk->resolution - 1);
        i = 0;
        index = 0;
        for(; i < numElements; ++i) {
            index = i << 2;
            fprintf(
                f,
                "f %u/%u %u/%u %u/%u %u/%u\n",
                chunk->indices[index + 0], chunk->indices[index + 0],
                chunk->indices[index + 1], chunk->indices[index + 1],
                chunk->indices[index + 2], chunk->indices[index + 2],
                chunk->indices[index + 3], chunk->indices[index + 3]
            );
        }

        logPrintf("Generated and saved terraing chunk: \'%s\'\n", folderpath);

        // cleanup
        fclose(f);
        delete(chunk->vertices);
        delete(chunk->texcoords);
        delete(chunk->indices);
        delete(chunk->normalHeightMapPixels);
        delete(chunk);

        ++chunkCount;
    }

    return 0;
}

